from .pairings import Result, Paired, Unpaired, Pairing, RoundPairings, GroupPairings, TournamentPairings
from .ids import GameId, GroupId, RoundId, gameId, roundId, groupId, stringifyId
from .mapping import GamesMapping
from .sources import ChessResultsPairings, PairingsSource

__all__ = [
  'Result', 'Paired', 'Unpaired', 'Pairing', 'RoundPairings', 'GroupPairings', 'TournamentPairings',
  'GameId', 'GroupId', 'RoundId', 'gameId', 'roundId', 'groupId', 'GamesMapping', 'stringifyId',
  'ChessResultsPairings', 'PairingsSource',
]


from typing_extensions import TypedDict

class GroupId(TypedDict):
  tournId: str
  group: str

class RoundId(GroupId):
  round: str

class GameId(RoundId):
  board: str

def gameId(tournId: str, group: str, round: str, board: str) -> GameId:
  return GameId(tournId=tournId, group=group, round=round, board=board)

def stringifyId(tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId}/{group}/{round}/{board}'

def roundId(tournId: str, group: str, round: str) -> RoundId:
  return RoundId(tournId=tournId, group=group, round=round)

def groupId(tournId: str, group: str) -> GroupId:
  return GroupId(tournId=tournId, group=group)



from typing import Callable, Mapping, Sequence, Iterable, Literal, overload
import chess.pgn
from chess_pairings.types import GamesMapping, gameId

@overload
def classify(
  games: Iterable[chess.pgn.Game], *,
  headers2tournId: Callable[[Mapping[str, str]], str | None] = lambda x: x.get('Event'),
  headers2group: Callable[[Mapping[str, str]], str | None],
  return_unclassified: Literal[True],
) -> tuple[GamesMapping[chess.pgn.Game], Sequence[chess.pgn.Game]]:
  ...

@overload
def classify(
  games: Iterable[chess.pgn.Game], *,
  headers2tournId: Callable[[Mapping[str, str]], str | None] = lambda x: x.get('Event'),
  headers2group: Callable[[Mapping[str, str]], str | None],
  return_unclassified: Literal[False] = False,
) -> GamesMapping[chess.pgn.Game]:
  ...
  
def classify(
  games: Iterable[chess.pgn.Game], *,
  headers2tournId: Callable[[Mapping[str, str]], str | None] = lambda x: x.get('Event'),
  headers2group: Callable[[Mapping[str, str]], str | None],
  return_unclassified: bool = False,
) -> tuple[GamesMapping[chess.pgn.Game], Sequence[chess.pgn.Game]] | GamesMapping[chess.pgn.Game]:    
  
  classified_games = GamesMapping[chess.pgn.Game]()
  unclassified_games = []
  
  for game in games:
    hdrs = game.headers
    tournId = headers2tournId(hdrs)
    group = headers2group(hdrs)
    rnd_brd = hdrs.get('Round', '').split('.')
    if tournId is None or group is None or len(rnd_brd) != 2:
      unclassified_games.append(game)
    else:
      round, board = rnd_brd
      classified_games[gameId(tournId, group, round, board)] = game

  if return_unclassified:
    return classified_games, unclassified_games
  else:
    return classified_games


from typing import MutableMapping, Mapping, Generic, TypeVar, Iterable
from chess_pairings.types import GameId, GroupId, RoundId, gameId, groupId, roundId

T = TypeVar('T')

class GamesMapping(MutableMapping[GameId, T], Generic[T]):

  @classmethod
  def from_pairs(cls, pairs: Iterable[tuple[GameId, T]]) -> 'GamesMapping[T]':
    mapping = cls()
    for key, value in pairs:
      mapping[key] = value
    return mapping

  def __init__(self, init: Mapping[str, dict[str, dict[str, dict[str, T]]]] = {}):
    """Mapping tnmtId -> group -> round -> board -> T"""
    self.dict: dict[str, dict[str, dict[str, dict[str, T]]]] = dict(init)

  def _getitem(self, *, tournId: str, group: str, round: str, board: str) -> T:
    return self.dict[tournId][group][round][board]
  
  def __getitem__(self, key: GameId) -> T:
    return self._getitem(**key)

  def _setitem(self, *, tournId: str, group: str, round: str, board: str, value: T) -> None:
    if tournId not in self.dict:
      self.dict[tournId] = {}
    if group not in self.dict[tournId]:
      self.dict[tournId][group] = {}
    if round not in self.dict[tournId][group]:
      self.dict[tournId][group][round] = {}
    self.dict[tournId][group][round][board] = value

  def __setitem__(self, key: GameId, value: T) -> None:
    self._setitem(**key, value=value)

  def __delitem__(self, *, tournId: str, group: str, round: str, board: str) -> None:
    del self.dict[tournId][group][round][board]
    if not self.dict[tournId][group][round]:
      del self.dict[tournId][group][round]
    if not self.dict[tournId][group]:
      del self.dict[tournId][group]
    if not self.dict[tournId]:
      del self.dict[tournId]

  def __len__(self) -> int:
    return len(list(self.gameIds()))
  
  def __repr__(self) -> str:
    return f'GamesMapping({repr(self.dict)})'
  
  def tournIds(self) -> Iterable[str]:
    return self.dict.keys()
  
  def groupIds(self) -> Iterable[GroupId]:
    for tournId, groups in self.dict.items():
      for group in groups:
        yield groupId(tournId, group)

  def roundIds(self) -> Iterable[RoundId]:
    for tournId, groups in self.dict.items():
      for group, rounds in groups.items():
        for round in rounds:
          yield roundId(tournId, group, round)

  def gameIds(self) -> Iterable[GameId]:
    for tournId, groups in self.dict.items():
      for group, rounds in groups.items():
        for round, boards in rounds.items():
          for board in boards:
            yield gameId(tournId, group, round, board)

  def __iter__(self) -> Iterable[GameId]:
    return self.gameIds()  
